from pankoff.base import BaseValidator

init_template = "def __init__({arguments}):\n\t{assignments}"


def autoinit(klass=None, verbose=False):
    def inner(cls):
        attrs = ["self"]
        assignments = []
        namespace = {}
        for attr in vars(cls).values():
            if isinstance(attr, BaseValidator):
                attrs.append(attr.attr_name)
                assignments.append(f"self.{attr.attr_name} = {attr.attr_name}")
        init = init_template.format(
            arguments=", ".join(attrs),
            assignments="\n\t".join(assignments or ["pass"])
        )
        if verbose:
            print(init)
        exec(init, namespace)
        cls.__init__ = namespace["__init__"]
        return cls

    if klass is not None:
        return inner(klass)
    return inner



import functools
import inspect
from abc import ABC, abstractmethod

from pankoff.exceptions import ValidationError


# CAUTION!!! do not touch anything here


def _invalidate_call_cache(instance, target):
    instance_type = type(instance)
    for base in instance_type.mro():
        if hasattr(base, target):
            attribute = getattr(base, target, None)
            if hasattr(attribute, "__called__"):
                del attribute.__called__


class _Descriptor:

    def __init__(self, **kwargs):
        _invalidate_call_cache(self, target="__setup__")

    def __set_name__(self, owner, name):
        self.attr_name = name

    def __set__(self, instance, value, errors, **kwargs):
        _invalidate_call_cache(self, target="validate")
        if errors:
            raise ValidationError(errors)
        vars(instance)[self.attr_name] = value


class BaseValidator(ABC, _Descriptor):

    def __init_subclass__(cls, **kwargs):
        def __cached_wrapper(func):
            @functools.wraps(func)
            def __inner(*args, **kw):
                __inner.__called__ = True
                ret = func(*args, **kw)
                return ret

            __inner.__wrapped__ = func
            return __inner

        for func_name in ("__setup__", "validate"):
            if hasattr(cls, func_name):
                setattr(cls, func_name, __cached_wrapper(getattr(cls, func_name)))

    def __init__(self, __mro__=None, **kwargs):
        kw = {}
        if not __mro__:
            __mro__ = type(self).mro()
        base, __mro__ = __mro__[0], __mro__[1:]

        signature = inspect.signature(base.__setup__)
        for parameter in signature.parameters.values():
            if parameter.name in kwargs:
                kw[parameter.name] = kwargs.pop(parameter.name)
        if base.__setup__ is not BaseValidator.__setup__:
            if not getattr(base.__setup__, "__called__", False):
                base.__setup__(self, **kw)
        super(base, self).__init__(__mro__=__mro__, **kwargs)

    def __set__(self, instance, value, __mro__=None, errors=None):
        errors = [] if errors is None else errors
        if not __mro__:
            __mro__ = type(self).mro()
        base, __mro__ = __mro__[0], __mro__[1:]
        if base.validate is not BaseValidator.validate:
            if not getattr(base.validate, "__called__", False):
                try:
                    base.validate(self, instance, value)
                except ValidationError as exc:
                    errors.append(str(exc))
        super(base, self).__set__(instance, value, __mro__=__mro__, errors=errors)

    def __setup__(self, *args, **kwargs):
        return NotImplemented

    @abstractmethod
    def validate(self, instance, value):
        return NotImplemented



import numbers

from pankoff.base import BaseValidator
from pankoff.exceptions import ValidationError


class Sized(BaseValidator):

    def __setup__(self, min_size=None, max_size=None):
        self.min_size = min_size
        self.max_size = max_size

    def validate(self, instance, value):
        if self.min_size is not None and len(value) < self.min_size:
            raise ValidationError(f"Attribute `{self.attr_name}` length should be >= {self.min_size}")
        elif self.max_size is not None and len(value) > self.max_size:
            raise ValidationError(f"Attribute `{self.attr_name}` length should be <= {self.max_size}")


class Type(BaseValidator):

    def __setup__(self, types):
        self.types = types

    def validate(self, instance, value):
        if not isinstance(value, self.types):
            raise ValidationError(f"Attribute `{self.attr_name}` should be an instance of `{self.types}`")


class String(Type):

    def __setup__(self, types=(str,)):
        super(String, self).__setup__(types)


class Number(Type):

    def __setup__(self, min_value=None, max_value=None, types=(numbers.Number,)):
        super(Number, self).__setup__(types)
        self.min_value = min_value
        self.max_value = max_value

    def validate(self, instance, value):
        super(Number, self).validate(instance, value)
        if self.min_value is not None and value < self.min_value:
            raise ValidationError(f"Attribute `{self.attr_name}` should be >= {self.min_value}")
        elif self.max_value is not None and value > self.max_value:
            raise ValidationError(f"Attribute `{self.attr_name}` should be <= {self.max_value}")



from distutils.core import setup

version = "3.0"
setup(
    name='pankoff',  # How you named your package folder (MyLib)
    packages=['pankoff'],  # Chose the same as "name"
    version=version,  # Start with a small number and increase it with every change you make
    license='MIT',  # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    description='Easy fields validator',  # Give a short description about your library
    author='Yaroslav Pankovych',  # Type in your name
    author_email='flower.moor@gmail.com',  # Type in your E-Mail
    url='https://github.com/ypankovych/pankoff',  # Provide either the link to your github or to your website
    download_url=f'https://github.com/ypankovych/pankoff/archive/refs/tags/{version}.tar.gz',  # I explain this later on
    keywords=['validation', 'easy', 'pure python'],  # Keywords that define your package best
    classifiers=[
        'Development Status :: 3 - Alpha',
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',  # Define that your audience are developers
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',  # Again, pick a license
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)



def combine(*validators, **kwargs):
    """
    Returns a combination of validators. E.g
    >>> name_validator = combine(Type, Sized)
    >>> name_validator(types=(str,), min_size=10)
    Of alternativelly:
    >>> name_validator = combine(Type, Sized, types=(str,), min_size=10)
    """
    def __repr__(self):
        validator_names = ", ".join(validator.__name__ for validator in type(self).__bases__)
        return f"{type(self).__name__}({validator_names})"

    klass = type(
        "CombinedValidator",
        validators,
        {"__repr__": __repr__}
    )
    klass._validators = klass.__bases__
    if not kwargs:
        return klass
    return klass(**kwargs)
